/***************************************************************************
                          addpal.c  -  description
                             -------------------
    begin                : Sat Jul 8 2000
    copyright            : (C) 2000 by Simon E. Reichelt
    email                : s.reichelt@gmx.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define PAL_SIZE 768


typedef struct
{
	char		identification[4];
	int			numlumps;
	int			infotableofs;
} wadinfo_t;


typedef struct
{
	int			filepos;
	int			disksize;
	int			size;
	char		type;
	char		compression;
	char		pad1, pad2;
	char		name[16];
} lumpinfo_t;


typedef unsigned char byte_t;


FILE *OpenFile(char f_name[], char mode[])
{
	FILE *file_p;
	
	file_p = fopen(f_name, mode);
	if (!file_p)
	{
		fprintf(stderr, "Couldn't open file \"%s\".\n",
		        f_name);
		exit(2);
	}
	
	return file_p;
}


int GetFileSize(FILE *file_p)
{
	int f_size;
	
	fseek(file_p, 0, SEEK_END);
	
	f_size = ftell(file_p);
	
	rewind(file_p);
	
	return f_size;
}


void CheckWadFile(FILE *wad_f)
{
	wadinfo_t w_info;
	lumpinfo_t pal_lmp;
	
	fread(&w_info, sizeof (wadinfo_t), 1, wad_f);
	
	if (strncmp(w_info.identification, "WAD2", 4))
	{
		fprintf(stderr, "WAD2 file is invalid.\n");
		exit(3);
	}
	
	fseek(wad_f, w_info.infotableofs, SEEK_SET);
	fread(&pal_lmp, sizeof (lumpinfo_t), 1, wad_f);
	
	if (!strcmp(pal_lmp.name, "PALETTE"))
	{
		fprintf(stderr, "WAD2 file already owns"
		                " a palette lump.\n");		
		fprintf(stderr, "Aborting...\n");
		exit(5);
	}
	
	rewind(wad_f);
}


void CheckPalette(FILE *pal_f)
{
	if (GetFileSize(pal_f) != PAL_SIZE)
	{
		fprintf(stderr, "LMP file is invalid.\n");
		exit(4);
	}
}


void CheckFiles(FILE *wad_f, FILE *pal_f)
{			
	CheckWadFile(wad_f);
	CheckPalette(pal_f);
}

		
byte_t *CacheDiskFile(FILE *file_p, int *byte_cnt)
{
	int i = 0;
	char ch;
	byte_t *mem_block;
	
	mem_block = malloc(GetFileSize(file_p) * sizeof (byte_t));
	if (!mem_block)
	{
		fprintf(stderr, "Ran out of memory.\n");
		exit(6);
	}
	
	while ((ch = fgetc(file_p)) != EOF)
		mem_block[i++] = ch;
		
	rewind(file_p);
	
	if (byte_cnt != NULL)
		*byte_cnt = i;
		
	return mem_block;
}


void AddPalette(char w_name[], FILE *wad_f, FILE *pal_f)
{
	byte_t *mem_pal;
	byte_t *wad_mem, *wad_run;
	int byte_cnt, mem_offs;
	wadinfo_t w_info;
	lumpinfo_t pal_lmp;
	
	printf("Re-assembling WAD2 file...");

	mem_pal = CacheDiskFile(pal_f, &byte_cnt);
	wad_mem = CacheDiskFile(wad_f, &byte_cnt);
	
	wad_run = wad_mem;
	
	fclose(wad_f);
	remove(w_name);
	
	wad_f = OpenFile(w_name, "w");
	
	memcpy(&w_info, wad_mem, sizeof (wadinfo_t));
	
	w_info.infotableofs += PAL_SIZE;
	
	fwrite(&w_info, sizeof (wadinfo_t), 1, wad_f);
	wad_run += sizeof (wadinfo_t);
	
	fwrite(wad_run, w_info.infotableofs - PAL_SIZE
	       - sizeof (wadinfo_t),
	       1, wad_f);
	fwrite(mem_pal, PAL_SIZE, 1, wad_f);
	
	strcpy(pal_lmp.name, "PALETTE");
	pal_lmp.filepos = w_info.infotableofs - PAL_SIZE;
	fwrite(&pal_lmp, sizeof (lumpinfo_t), 1, wad_f);
	
	mem_offs = w_info.infotableofs - PAL_SIZE;
	
	wad_run = wad_mem + mem_offs;
	
	fwrite(wad_run, byte_cnt - mem_offs, 1, wad_f);
	
	printf(" done.\n");
	printf("Added palette lump at %d bytes.\n",
	       pal_lmp.filepos);
}	
	

void Sub_AddPalette(int a_cnt, char *a_vec[])
{
	FILE *w_file, *p_file;
	
	if (a_cnt != 4)
	{
		fprintf(stderr, "\naddpal: palwiz -addpal <wadfile>"
		                " <lmpfile>\n\n");
		exit(8);
	}
	
	w_file = OpenFile(a_vec[2], "r");
	
	p_file = OpenFile(a_vec[3], "r");
	
	CheckFiles(w_file, p_file);
	
	AddPalette(a_vec[2], w_file, p_file);
	
	fclose(w_file);
	fclose(p_file);
}


void GenerateAscFile(char lmp_name[], FILE *lmp_f)
{
	FILE *asc_f;
	byte_t *mem_p;
	byte_t clr_chans[3];
	int i;
	char *f_name;
	
	f_name = strdup(lmp_name);
	
	f_name[strlen(lmp_name) - 4] = '\0';
	strcat(f_name, ".asc");
	
	asc_f = OpenFile(f_name, "w");
	mem_p = CacheDiskFile(lmp_f, NULL);
	
	fprintf(asc_f, "\"%s\"\n{\n", lmp_name);
	
	for (i = 0; i < PAL_SIZE / 3; i++)
	{
		memcpy(clr_chans, mem_p, sizeof (byte_t) * 3);
		
		fprintf(asc_f, "\t%d  %d  %d\n",
		        clr_chans[0], clr_chans[1], clr_chans[2]);
		
		mem_p += 3;
	}
	
	fputc('}', asc_f);
	fclose(asc_f);
	
	printf("Wrote ASCII colormap \"%s\".\n", f_name);
}


void Sub_LumpToASCII(int a_cnt, char *a_vec[])
{
	FILE *lmp_f;
	
	if (a_cnt != 3)
	{
		fprintf(stderr, "\nlmp2asc: palwiz -lmp2asc <lmpfile>\n\n");
		exit(7);
	}
	
	lmp_f = OpenFile(a_vec[2], "r");
	
	GenerateAscFile(a_vec[2], lmp_f);
	
	fclose(lmp_f);
}


void InitPalScript(FILE *asc_f, char lmp_name[])
{
	int i;
	char ch;

	ch = fgetc(asc_f);

	if (ch != '\"')
	{
		fprintf(stderr, "ERROR: File may not be "
		                "a ASC file.\n");
		exit(11);
	}
	
	for (i = 0; (ch = fgetc(asc_f)) != EOF; i++)
	{
		if (ch == '\"')
			break;
			
		lmp_name[i] = ch;
	}
	
	lmp_name[i] = '\0';
}


void GeneratePalData(FILE *asc_f)
{
	char ch;
	char lmp_name[256];
	byte_t clr_chans[3];
	int i;
	FILE *lmp_f;
	
	InitPalScript(asc_f, lmp_name);	
	
	lmp_f = OpenFile(lmp_name, "w");
	
	ch = fgetc(asc_f);
	ch = fgetc(asc_f);
	if (ch != '{')
	{
		fprintf(stderr, "ERROR: File misses "
		        "opening brace.\n");
		exit(12);
	}
	
	for (i = 0; i < PAL_SIZE / 3; i++)
	{
		fscanf(asc_f, "%d %d %d", (int*) &clr_chans[0],
		       (int*) &clr_chans[1], (int*) &clr_chans[2]);
		
		fwrite(clr_chans, sizeof (byte_t), 3, lmp_f);
	}
	
	ch = fgetc(asc_f);
	ch = fgetc(asc_f);
	if (ch != '}')
	{
		fprintf(stderr, "ERROR: Palette has more than %d"
		                " entries\n", PAL_SIZE / 3);
		exit(13);
	}
	
	fclose(lmp_f);
	
	printf("Wrote palette lump \"%s\".\n", lmp_name);
}
	
	
void Sub_ASCIIToLump(int a_cnt, char *a_vec[])
{
	FILE *asc_f;
	
	if (a_cnt != 3)
	{
		fprintf(stderr, "\nasc2lmp: palwiz -asc2lmp <ascfile>\n\n");
		exit(9);
	}
	
	asc_f = OpenFile(a_vec[2], "r");
	
	GeneratePalData(asc_f);
	
	fclose(asc_f);
}


void Sub_AboutInfo(void)
{
	printf("\n---------------------\n"
	       "   PalWiz v1.0\n"
	       "---------------------\n\n"
	       "Copyright (C) 2000 by\n"
	       "  Simon E. Reichelt\n\n");
}

	
int main(int argc, char *argv[])
{
	if (argc < 2)
	{
		fprintf(stderr, "\nusage: palwiz <mode> [...]\n\n");
		return 1;
	}
	
	if (!strcmp(argv[1], "-addpal"))
		Sub_AddPalette(argc, argv);
	else if (!strcmp(argv[1], "-lmp2asc"))
		Sub_LumpToASCII(argc, argv);
	else if (!strcmp(argv[1], "-asc2lmp"))
		Sub_ASCIIToLump(argc, argv);
	else if (!strcmp(argv[1], "-about"))
		Sub_AboutInfo();
	else
	{
		fprintf(stderr, "Unknown option \"%s\"\n",
		        argv[1]);
		return 0;
	}
	
	return 0;
}