/***************************************************************************
                          wadfile.c  -  description
                             -------------------
    begin                : Wed Jul 26 2000
    copyright            : (C) 2000 by Simon E. Reichelt
    email                : s.reichelt@gmx.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include <string.h>
#include <malloc.h>

#include "wadview.h"
#include "callbacks.h"


int wad_type;


int CheckMemory(void *mem_p)
{
	if (!mem_p)
	{
		WVError("Ran Out of Memory");
		return BAD_MEM;
	}
	else
		return 1;
}


FILE *OpenFile(char name[], char mode[])
{
	FILE *file;
	
	file = fopen(name, mode);
	if (!file)
	{
		WVError("Couldn't Open File.");		
		return NULL;
	}
	
	return file;
}


int CheckFileType(FILE *w_file)
{	
	wadinfo_t w_info;
	lumpinfo_t pal_lmp;
	
	fread(&w_info, sizeof (wadinfo_t), 1, w_file);
	
	if (!strncmp(w_info.identification, "WAD2", 4))
		wad_type = WAD2;
	else if (!strncmp(w_info.identification, "WAD3", 4))
		wad_type = WAD3;
	else
	{
		WVError("File Is Not A WAD File!");
		rewind(w_file);
		return -1;
	}
	
	fseek(w_file, w_info.infotableofs, SEEK_SET);
	fread(&pal_lmp, sizeof (lumpinfo_t), 1, w_file);
	
	if (wad_type == WAD2)
		if (strcmp(pal_lmp.name, "PALETTE"))
		{
			WVError("File Has No Palette Entry.");
			rewind(w_file);
			return -1;
		}
	
	rewind(w_file);
	
	return w_info.numlumps;
}


byte_t *LoadFileFromDisk(FILE *file_p)
{
	int i;
	int f_size;
	byte_t *mem_block;
	
	//Load a file from disk into virtual memory
	//so we can process it A LOT faster ;-)
	
	fseek(file_p, 0, SEEK_END);
	f_size = ftell(file_p);
	rewind(file_p);
	
	mem_block = malloc(f_size * sizeof (byte_t));
	
	for (i = 0; i < f_size; i++)
		mem_block[i] = fgetc(file_p);
		
	rewind(file_p);
	
	return mem_block;
}


void LoadWadFile(GtkWidget *widget, char f_name[])
{
	FILE *w_file;
	int lmp_cnt;
	texrep_t *textures;
	GdkGC *raw_gc;
	
	//This is the actual texture WAD
	//loading procedure
	
	w_file = OpenFile(f_name, "r");
	if (!w_file)
		return;
		
	ConPrint("Loading: %s\n", f_name);
		
	lmp_cnt = CheckFileType(w_file);
	if (lmp_cnt == -1)
	{
		ConPrint("LoadWadFile(): failure.\n");
		return;
	}
	
	FlushTextures((gpointer) widget, NULL);
	
	tex_cnt = lmp_cnt;
	
	raw_gc = gdk_gc_new(con_text->window);
	
	textures = CacheWadTextures(w_file, raw_gc);
	if (!textures)
	{
		ConPrint("LoadWadFile(): failure.\n");
		return;
	}
	
	SortCacheTextures(textures, lmp_cnt);
	
	tex_dat = textures;
	
	SetMainWinTitle(widget, f_name);	
	
	PopulateTexBrowser(textures, lmp_cnt, widget);	
	
	ConPrint("LoadWadFile(): success.\n");
}